import numpy, math, sys, os

def propTsTv(seq1, seq2):
  if len(seq1) != len(seq2):
    sys.stderr.write('Error! Sequences different lengths!\n')
    sys.exit(1)
    
  nchar = len(seq1)
  nTs = 0
  nTv = 0
  
  seq1 = seq1.lower()
  seq2 = seq2.lower()
  
  for i in xrange(nchar):
    if seq1[i] != seq2[i]:
      if seq1[i] == 'c' or seq1[i] == 't':
        if seq2[i] == 'a' or seq2[i] == 'g':
          #print 'transversion:',
          nTv += 1
        else:
          #print 'transition:', 
          nTs += 1
      else:
        if seq2[i] == 'a' or seq2[i] == 'g':
          #print 'transition:',
          nTs += 1
        else:
          #print 'transversion:',
          nTv += 1
      #print seq1[i], '->', seq2[i]
          
  return float(nTs)/nchar, float(nTv)/nchar

def countsubs(seq1, seq2):
  
  indices = {'a':0, 't':1, 'c':2, 'g':3}
  
  nchar = len(seq1)
  if len(seq2) != nchar:
    sys.stderr.write('Sequence lengths differ!\n')
    sys.exit(1)
  
  submat = numpy.matrix([0. for i in xrange(16)]).reshape(4,4)
  
  for i in xrange(nchar):
    submat[indices[seq1[i]], indices[seq2[i]]] += 1
  
  return submat

# Tamura-Nei distance
# Seems to be wrong!
def tn93(align):

  ntaxa = align.dim
  
  dmat = numpy.matrix([0. for i in xrange (ntaxa ** 2)]).reshape((ntaxa, ntaxa))
  
  nchar = 0
  for i in xrange(ntaxa):
    seqI = align.sequences[i].sequence
    if not nchar:
      nchar = len(seqI)
    #nAi = seqI.count('a')
    #nTi = seqI.count('t')
    #nCi = seqI.count('c')
    #nGi = seqI.count('g')
    for j in xrange(i):
      seqJ = align.sequences[j].sequence
      
      subs = countsubs(seqI, seqJ)/nchar
      rowsums = numpy.sum(subs, axis=0)
      colsums = numpy.sum(subs, axis=1)
      
      gA = rowsums[0,0] + colsums[0,0] - subs[0,0]
      gT = rowsums[0,1] + colsums[1,0] - subs[1,1]
      gC = rowsums[0,2] + colsums[2,0] - subs[2,2]
      gG = rowsums[0,3] + colsums[3,0] - subs[3,3]
      
      p1 = subs[0,3] + subs[3,0]# AG transitions
      print 'purine transitions:', p1
      p2 = subs[1,2] + subs[2,1]# TC transitions
      print 'pyrimidine transitions:', p2
      q = subs[0,1] + subs[0,2] + subs[1,0] + subs[1,3] + subs[2,0] + subs[2,3] + subs[3,1] + subs[3,2]
      print 'transversions:', q
      
      #gA = (nAi + seqJ.count('a')) / (2 * nchar)
      #gT = (nTi + seqJ.count('t')) / (2 * nchar)
      #gC = (nCi + seqJ.count('c')) / (2 * nchar)
      #gG = (nGi + seqJ.count('g')) / (2 * nchar)
      
      
      
      k1 = 2 * gA * gG / (gA + gG)
      k2 = 2 * gT * gC / (gT + gC)
      k3 = 2 * ((gA + gG) * (gT + gC) -(gA * gG * (gT + gC)) / (gA + gG) - (gT * gC * (gA + gG) / (gT + gC))) 
      k4 = 2 * (gA * gG + gT * gC + (gA + gG) * (gT + gC))
      
      w1 = 1 - p1/k1 - q/(gA + gG)
      w2 = 1 - p2/k2 - q/(gT + gC)
      w3 = 1 - q/(2 * (gA + gG) * (gT + gC))
      
      dmat[i,j] = -k1 * math.log(w1) - k2 * math.log(w2) - k3 * math.log(w3)
      
  return dmat
  
def k2p(align):
  
  ntaxa = align.dim
  
  dmat = numpy.matrix([0. for i in xrange (ntaxa ** 2)]).reshape((ntaxa, ntaxa))
  
  psum = 0
  qsum = 0
  
  for i in xrange(ntaxa):
    for j in xrange(i):
      p, q = propTsTv(align.sequences[i].sequence, align.sequences[j].sequence)
      
      if p == 0:
        p = 1e-6
        #sys.stderr.write('Very low Ts. Setting to %f.\n' % p)
      if q >= (1 - 2 * p):
        #print 'q:', q, 'p:', p
        q = 1 - 2 * p - 1e-6
        #sys.stderr.write('Impossible Ts:Tv ratio. Setting Tv to %f.\n' % q)
        #sys.exit(1)
      elif q >= 0.5:
        #print 'q:', q, 'p:', p
        q = 0.5 - 1e-6
        #sys.stderr.write('Impossible Tv proportion. Setting to %f.\n' % q)
        #sys.exit(1)
      psum += p
      qsum += q
      try:
        dmat[i,j] = -0.5 * math.log(1 - 2 * p -q) - 0.25 * math.log(1 - 2 * q)
        dmat[j,i] = dmat[i,j]
      except IndexError:
        sys.stderr.write('i: %d, j: %d, dmat shape: %s\n' % (i, j, str(dmat.shape)))
      except ValueError:
        print 'p:', p
        print 'q:', q
        raise
    
  #printmat(dmat)
  
  return dmat

def pdist(seqI, seqJ):
  
  nchar = len(seqI)
  if (len(seqJ) != nchar):
    sys.stderr.write('Sequences are not of equal length!\n')
    exit(1)
  
  ndiff = 0  
  for i in xrange(nchar):
    if (seqI[i] != seqJ[i]):
      ndiff += 1
      
  return float(ndiff) / nchar

def jc(align):
  ntaxa = align.nTax
  #print 'ntaxa:', ntaxa
  dmat = numpy.matrix([0. for i in xrange (ntaxa ** 2)]).reshape((ntaxa, ntaxa))
  
  for i in xrange(ntaxa):
    for j in xrange(i):
      p = pdist(align.sequences[i].sequence, align.sequences[j].sequence)
      try:
        dmat[i,j] = -3 * math.log(1 - 4 * p / 3) / 4 
      except ValueError:
        ##################################
        #                                #
        #  DO SOMETHING ABOUT THIS!!!!!  #
        #                                #
        ##################################
        dmat[i,j] = 10.
      dmat[j,i] = dmat[i,j]
      
  return dmat
     
def writemat(mat, filename, taxnames=[]):
  outfile = open(filename, 'w')
  nrow, ncol = mat.shape
  
  if nrow != ncol:
    sys.stderr.write('Matrix has different numbers of rows and columns!')
    sys.exit(1)
    
  outfile.write(' %d\n' % nrow)
  
  for i in xrange(nrow):
    if taxnames:
      outfile.write('%-10s' % taxnames[i][:10])
    else:
      outfile.write('tax%-7d' % i) 
    for j in xrange(ncol):
      outfile.write(' %.6g' % mat[i,j])
    outfile.write('\n')
    


  