import numpy, math, sys, types

"""treetuple must be a 1-tuple containing the tree list!"""
def treelist2string(treetuple, namelist, first=True):
  treestr = ''
  
  if first:
    node = treetuple[0]
  
  else:
    if len(treetuple) != 2:
      sys.stderr.write('Odd, should be a 2-tuple: %s\n' % str(treetuple))
      sys.exit(1)
    
    node, brlen = treetuple
  
  
  if type(node) == types.ListType:
    treestr += '('
    for i in xrange(len(node)):
      if i > 0:
        treestr += ','
      treestr += treelist2string(node[i], namelist, False)
    if first:
      treestr += ');'
    else:
      treestr += '):%f' % brlen
    
  else:
    if first:
      try:
        treestr += '%s;' % namelist[node]
      except TypeError:
        sys.stderr.write('Problem making tree string. Try passing the treelist as a tuple: (treelist,)\n')
        raise
    else:
      treestr += '%s:%f' % (namelist[node], brlen)  
  
  return treestr


def dnew(dmat, tax1, tax2):
  ntaxa = dmat.shape[0]
  dist = 0.5 * dmat[tax1, tax2]
  
  summation = 0.
  for i in xrange(ntaxa):
    summation += dmat[tax1, i]
    summation -= dmat[tax2, i]
  
  return dist + (1.0/(2 * (ntaxa - 2)) * summation)
  
def njreduce(dmat, tax1, tax2):
  
  nrow, ncol = dmat.shape
    
  if nrow != ncol:
    sys.stderr.write('Distance matrix has different numbers of rows and cols!\n')
    sys.exit(1)
  
  if tax2 >= tax1:
    sys.stderr.write('ID for tax2 >= ID for tax1; apparently matrix not lower triangular?\n')
    sys.exit(1)
  
  dfu = dnew(dmat, tax1, tax2)
  dgu = dnew(dmat, tax2, tax1)
  
  newmat = numpy.matrix([0. for i in xrange((nrow - 1) ** 2)]).reshape((nrow - 1), (nrow - 1))
  
  for i in xrange(nrow - 1):
    for j in xrange(i):
      if i == tax2:
        if j >= tax1:
          newmat[i,j] = 0.5 * (dmat[tax1,j+1] - dfu) + 0.5 * (dmat[tax2,j+1] - dgu)
        else:
          newmat[i,j] = 0.5 * (dmat[tax1,j] - dfu) + 0.5 * (dmat[tax2,j] - dgu) 
      elif j == tax2:
        if i >= tax1:
          newmat[i,j] = 0.5 * (dmat[i+1,tax1] - dfu) + 0.5 * (dmat[i+1,tax2] - dgu)
        else:
          newmat[i,j] = 0.5 * (dmat[i,tax1] - dfu) + 0.5 * (dmat[i,tax2] - dgu)
        
      elif i >= tax1:
        if j >= tax1:
          newmat[i,j] = dmat[i + 1, j + 1]
        else:
          newmat[i,j] = dmat[i + 1, j]
      else:
        if j >= tax1:
          sys.stderr.write('This should never happen... looks like i is less than j\n')
          sys.exit(1)
          newmat[i,j] = dmat[i, j + 1]
        else:
          newmat[i,j] = dmat[i,j]
          
      newmat[j,i] = newmat[i,j]

  return dfu, dgu, newmat

def nj(dmat):
    
  ntaxa = dmat.shape[0]
  
  tree = [(i,0.) for i in xrange(ntaxa)]
    
  #njoined = 1
  
  while len(tree) > 2:#njoined < ntaxa:
    Qmat = numpy.matrix([0. for i in xrange(ntaxa ** 2)]).reshape(dmat.shape)
    minval = 1
    minpair = (-1,-1)
    for i in xrange(ntaxa):
      for j in xrange(i):
        Qmat[i,j] = (ntaxa - 2) * dmat[i,j]
        for k in xrange(ntaxa):
          Qmat[i,j] -= dmat[i,k]
          Qmat[i,j] -= dmat[j,k]
          
        if minpair[0] < 0 or Qmat[i,j] < minval:
          minval = Qmat[i,j]
          minpair = (i,j)
    
    dfu, dgu, dmat = njreduce(dmat, minpair[0], minpair[1])
    
    #print 'Species %d (%f) joins species %d (%f)' % (minpair[0], dfu, minpair[1], dgu)
    #print 'Qmat:'
    #printmat(Qmat)
    
    tree[minpair[0]] = (tree[minpair[0]][0], dfu)
    tree[minpair[1]] = (tree[minpair[1]][0], dgu)
    
    tree[minpair[1]] = ([tree[minpair[0]], tree[minpair[1]]],0.)
    del tree[minpair[0]]
        
    ntaxa -= 1
    #njoined += 1
    
  if len(tree) != 2:
    sys.stderr.write('Tree has wrong number of root children: %d\n' % len(tree))
    sys.exit(1)
    
  rootd = dmat[1,0]/2
  tree[0] = (tree[0][0], rootd)
  tree[1] = (tree[1][0], rootd)
  
  return tree

def upgma(dmat):
  ntaxa = dmat.shape[0]
  
  tree = [(i,0.) for i in xrange(ntaxa)]
    
  #njoined = 1
    
  while len(tree) > 1:#njoined < ntaxa:
    minval = 1
    minpair = (-1,-1)
    for i in xrange(ntaxa):
      for j in xrange(i):    
        if minpair[0] < 0 or dmat[i,j] < minval:
          minval = dmat[i,j]
          minpair = (i,j)

    brlen, dmat = upgmareduce(dmat, minpair[0], minpair[1])
    
    tree[minpair[0]] = (tree[minpair[0]][0], brlen/2 - disttotip(tree[minpair[0]]))
    tree[minpair[1]] = (tree[minpair[1]][0], brlen/2 - disttotip(tree[minpair[1]]))
    tree[minpair[1]] = ([tree[minpair[0]], tree[minpair[1]]],0.)
    del tree[minpair[0]]
        
    ntaxa -= 1
    #njoined += 1

  #if len(tree) != 2:
  #  sys.stderr.write('Tree has wrong number of root children: %d\n' % len(tree))
  #  sys.exit(1)
  
  #rootd = dmat[1,0]/2
  #tree[0] = (tree[0][0], rootd)
  #tree[1] = (tree[1][0], rootd)
  
  tree = tree[0][0]
    
  return tree
    
   
def upgmareduce(dmat, tax1, tax2):
  
  nrow, ncol = dmat.shape
    
  if nrow != ncol:
    sys.stderr.write('Distance matrix has different numbers of rows and cols!\n')
    sys.exit(1)
  
  if tax2 >= tax1:
    sys.stderr.write('ID for tax2 >= ID for tax1; apparently matrix not lower triangular?\n')
    sys.exit(1)
  
  d12 = dmat[tax1, tax2]
    
  newmat = numpy.matrix([0. for i in xrange((nrow - 1) ** 2)]).reshape((nrow - 1), (nrow - 1))
  
  for i in xrange(nrow - 1):
    for j in xrange(i):
      if i == tax2:
        if j >= tax1:
          newmat[i,j] = 0.5 * (dmat[tax1,j+1] + dmat[tax2,j+1])
        else:
          newmat[i,j] = 0.5 * (dmat[tax1,j] + dmat[tax2,j]) 
      elif j == tax2:
        if i >= tax1:
          newmat[i,j] = 0.5 * (dmat[i+1,tax1] + dmat[i+1,tax2])
        else:
          newmat[i,j] = 0.5 * (dmat[i,tax1] + dmat[i,tax2])
        
      elif i >= tax1:
        if j >= tax1:
          newmat[i,j] = dmat[i + 1, j + 1]
        else:
          newmat[i,j] = dmat[i + 1, j]
      else:
        if j >= tax1:
          sys.stderr.write('This should never happen... looks like i is less than j\n')
          sys.exit(1)
          newmat[i,j] = dmat[i, j + 1]
        else:
          newmat[i,j] = dmat[i,j]
          
      newmat[j,i] = newmat[i,j]

  return d12, newmat

def disttotip(subtree):
  if type(subtree[0]) == types.ListType: 
    return subtree[1] + disttotip(subtree[0][0])
  else:
    return subtree[1]